/** @ MilleTool.cpp 
 *
 *  Interface to Millepede
 *
 */

#include "MilleTool.h"
#include "constants.h"
#include "Track.h"
#include <iostream>

using namespace std;

bool MilleTool::initialize(const Planes& geom){

  // init phase

  m_generator = new TRandom();

  m_geometry = geom;
  Xmeas.resize(m_geometry.size());
  xTrue.resize(m_geometry.size());  
  Ymeas.resize(m_geometry.size());
  yTrue.resize(m_geometry.size());  
 
  // Global parameters vectors
  //
  // !!! All is dimensionned as if we want to find all the 6DOFs
  // But we will inform Millepede of the DOFs to take into account

  bool   dofs[6];
  double Sigm[6];
  double res_cut, res_cut_init, start_chi_cut;

  res_cut       = m_res_cut; 
  res_cut_init  = m_res_cut_init;
  start_chi_cut = m_start_chi_cut;

  for (int i = 0; i<6; ++i)
  {
    dofs[i] = m_dofs[i];
    Sigm[i] = m_Sigm[i];
  }

  int pSize = m_geometry.size(); // Number of elements to align

  m_derGB.resize(6*pSize);
  m_derNonLin.resize(6*pSize);
  m_par.resize(6*pSize);
  m_error.resize(6*pSize);
  m_pull.resize(6*pSize);

  // Local
  m_derLC.resize(4); // One track is 2 straight lines, so 4 local params

  m_alignTool.InitMille(&dofs[0], &Sigm[0], m_geometry.size(), m_derLC.size(), 
			 start_chi_cut, 3, res_cut, res_cut_init);

  //
  // Constraint equation on the X, Y, and Z translations (prevent global translation and shearing)
  //

  // First of all the Global Translation

  std::vector<double> Const_TX, Const_TY, Const_TZ, Const_RX, Const_RY, Const_RZ;
  Const_TX.resize(6*pSize);
  Const_TY.resize(6*pSize);
  Const_TZ.resize(6*pSize);
  Const_RX.resize(6*pSize);
  Const_RY.resize(6*pSize);
  Const_RZ.resize(6*pSize);

  for (int i = 0; i< pSize; ++i)
  {
    Const_TX[i] = 0.0;
    Const_TY[i] = 0.0;
    Const_TZ[i] = 0.0;
    Const_RX[i] = 0.0;
    Const_RY[i] = 0.0;
    Const_RZ[i] = 0.0;
  }

  for (int i = 0; i< pSize; ++i)
  {
    Const_TX[i]         = 1.0;
    Const_TY[i+pSize]   = 1.0;
    Const_TZ[i+2*pSize] = 1.0;
    Const_RX[i+3*pSize] = 1.0;
    Const_RY[i+4*pSize] = 1.0;
    Const_RZ[i+5*pSize] = 1.0;
  }

  // Feed Millepede (only if requested)  

  if (dofs[0]) m_alignTool.ConstF(&Const_TX[0],  0.0);   
  if (dofs[1]) m_alignTool.ConstF(&Const_TY[0],  0.0); 
  if (dofs[2]) m_alignTool.ConstF(&Const_TZ[0],  0.0);
  if (dofs[3]) m_alignTool.ConstF(&Const_RX[0],  0.0);
  if (dofs[4]) m_alignTool.ConstF(&Const_RY[0],  0.0);
  if (dofs[5]) m_alignTool.ConstF(&Const_RZ[0],  0.0);

  // Then the shearings and scalings 

  double z_moy =0;
  double sigma_z_moy = 0;
  // z_average  
  for (unsigned int j=0; j<nStation; j++) {z_moy  += m_geometry[j].z();}
  z_moy /= nStation;
  // Error on z_moy
  for (unsigned int j=0; j<nStation; j++){
    sigma_z_moy += ( m_geometry[j].z()-z_moy)*( m_geometry[j].z()-z_moy);
  }
  sigma_z_moy /= nStation;
  
  // constraint on the shearing
  std::vector<double> Const_SX, Const_SY, Const_FZ;
  Const_SX.resize(6*pSize);
  Const_SY.resize(6*pSize);
  Const_FZ.resize(6*pSize);

  for (int i = 0; i< pSize; ++i)
  {
    Const_SX[i] = 0.0;
    Const_SY[i] = 0.0;
    Const_FZ[i] = 0.0;
  }

  for (int i = 0; i< pSize; ++i){
    Const_SX[i]         = (m_geometry[i].z()-z_moy)/sigma_z_moy;
    Const_SY[i+pSize]   = (m_geometry[i].z()-z_moy)/sigma_z_moy;
    Const_FZ[i+2*pSize] = (m_geometry[i].z()-z_moy)/sigma_z_moy;
  }

  // Feed Millepede (only if requested)  

  if (dofs[0]) m_alignTool.ConstF(&Const_SX[0],  0.0);   
  if (dofs[1]) m_alignTool.ConstF(&Const_SY[0],  0.0); 
  if (dofs[2]) m_alignTool.ConstF(&Const_FZ[0],  0.0);

  // this is it! Millepede is now ready to align...

  return true;
}

// add a track to millipede

bool MilleTool::addTrack(Track* trueTrack, 
                         std::vector<double> m_offsetX, 
                         std::vector<double> m_offsetY, 
                         std::vector<double> m_offsetZ, 
                         std::vector<double> m_offsetA, 
                         std::vector<double> m_offsetB, 
                         std::vector<double> m_offsetC, 
                         std::vector<double>& smearX,
                         std::vector<double>& smearY)
{


  //  cout << "Into addtrack" << endl;

  int nStation = m_geometry.size();

  for (int iStation = 0; iStation < nStation; ++iStation)
  { 
    Xmeas[iStation] = 0.;
    xTrue[iStation] = 0.;
    Ymeas[iStation] = 0.;
    yTrue[iStation] = 0.;
  }

  // Add the track measurements
  
  double track_orig[3];
  trueTrack->getOrigin(&track_orig[0]);
  double v_z = track_orig[2];
  bool   track_direction = trueTrack->getDirection();

  //  cout << endl;
  //  cout << "Track perf number = " << trueTrack->getTracknum() << endl;

  for (int iStation = 0; iStation < nStation; ++iStation)
  { 

    // Here we add the misalignment to the track parameters (this is a toy!)
    // Relation between derivatives and misalignment is described in note LHCb-2005-101

    double z_add = (m_generator->Rndm(1) - 0.5)* thickness;   // Sensor thickness effect

    // Perfect

    xTrue[iStation] = trueTrack->x(m_geometry[iStation].z()+z_add); 
    yTrue[iStation] = trueTrack->y(m_geometry[iStation].z()+z_add); 

    // Check if the track point is inside the sensor active area, and in a good sensor

    if (((m_geometry[iStation].z()+z_add) < v_z) && (track_direction == true)) continue;
    if (((m_geometry[iStation].z()+z_add) > v_z) && (track_direction == false)) continue;
    if (!(m_geometry[iStation].isInside(xTrue[iStation],yTrue[iStation]))) continue;

    // Misaligned

    Xmeas[iStation] = xTrue[iStation] + smearX[iStation]*resolution 
      - m_offsetX[iStation] 
      + yTrue[iStation]*m_offsetC[iStation]
      + trueTrack->tx()*
      ( m_offsetZ[iStation] 
	+ yTrue[iStation]*m_offsetA[iStation]
	+ xTrue[iStation]*m_offsetB[iStation]);

    Ymeas[iStation] = yTrue[iStation] + smearY[iStation]*resolution 
      - m_offsetY[iStation] 
      - xTrue[iStation]*m_offsetC[iStation]
      + trueTrack->ty()*
      ( m_offsetZ[iStation] 
	+ yTrue[iStation]*m_offsetA[iStation]
	+ xTrue[iStation]*m_offsetB[iStation]);
  }

  // Now we have the misaligned clusters. One could create the new track from that
  // one do here a simple least square fit

  double local_vecx[2];
  double local_matx[2][2];
  double local_vecy[2];
  double local_maty[2][2];
  
  double wghtx, rmeasx, detx;
  double wghty, rmeasy, dety;
  double z_cor;
  
  for (int i=0; i<2; i++) // reset local params
  {
    local_vecx[i] = 0.0;
    local_vecy[i] = 0.0;

    for (int j=0; j<2; j++)
    {
      local_matx[i][j] = 0.0;
      local_maty[i][j] = 0.0;
    }
  }

  for (int k=0; k<nStation; k++)
  {
    z_cor = m_geometry[k].z();
    wghtx = 1.;  // Assign the same error to all the clusters
    rmeasx = Xmeas[k];
    wghty = 1.;
    rmeasy = Ymeas[k];

    // Check that we are in the active area, and in the right VELO area

    if ((z_cor < v_z) && (track_direction == true)) continue;
    if ((z_cor > v_z) && (track_direction == false)) continue;
    if (!(m_geometry[k].isInside(rmeasx,rmeasy))) continue;
    
    local_vecx[0] += wghtx*rmeasx;
    local_vecx[1] += wghtx*rmeasx*z_cor;
    local_vecy[0] += wghty*rmeasy;
    local_vecy[1] += wghty*rmeasy*z_cor;
      
    local_matx[0][0] += wghtx;
    local_matx[1][0] += wghtx*z_cor;
    local_matx[1][1] += wghtx*z_cor*z_cor;
    local_maty[0][0] += wghty;
    local_maty[1][0] += wghty*z_cor;
    local_maty[1][1] += wghty*z_cor*z_cor;					
  }

  detx = 0.;
  dety = 0.;
    
  detx = local_matx[0][0]*local_matx[1][1]-local_matx[1][0]*local_matx[1][0];
  dety = local_maty[0][0]*local_maty[1][1]-local_maty[1][0]*local_maty[1][0];
    
  if (detx == 0. || dety == 0.) return false;
    
  double slope_x = (local_vecx[1]*local_matx[0][0]-local_vecx[0]*local_matx[1][0])/detx;
  double slope_y = (local_vecy[1]*local_maty[0][0]-local_vecy[0]*local_maty[1][0])/dety;
    
  double c_x     = (local_vecx[0]*local_matx[1][1]-local_vecx[1]*local_matx[1][0])/detx;
  double c_y     = (local_vecy[0]*local_maty[1][1]-local_vecy[1]*local_maty[1][0])/dety;
    
  Line new_X(slope_x,c_x);
  Line new_Y(slope_y,c_y);

  Track* misalignedTrack = new Track(new_X,new_Y); // Here is the misaligned track

  //
  // Then we add the track to Millepede
  //
    
  for (int iStation = 0; iStation < nStation; ++iStation)
  {

    if ((m_geometry[iStation].z() < v_z) && (track_direction == true)) continue;
    if ((m_geometry[iStation].z() > v_z) && (track_direction == false)) continue;
    if (!(m_geometry[iStation].isInside(Xmeas[iStation],Ymeas[iStation]))) continue;

    m_alignTool.ZerLoc(&m_derGB[0],&m_derLC[0],&m_derNonLin[0]); // Reset the derivatives arrays

    // Add the local derivatives for the X part of the track

    m_derLC[0]= 1.0;
    m_derLC[1] = m_geometry[iStation].z();
    m_derLC[2]= 0.0;
    m_derLC[3]= 0.0;

    // Then add the global one

    if (m_dofs[0]) m_derGB[iStation]                     = -1.0;
    if (m_dofs[1]) m_derGB[iStation+m_geometry.size()]   = 0.0;

    // Linear approximation
    //    if (m_dofs[2]) m_derGB[iStation+2*m_geometry.size()] = 1.*misalignedTrack.tx();
    //    if (m_dofs[3]) m_derGB[iStation+3*m_geometry.size()] = Ymeas[iStation]*misalignedTrack->tx();
    //    if (m_dofs[4]) m_derGB[iStation+4*m_geometry.size()] = Xmeas[iStation]*misalignedTrack->tx();

    // Non-Linear case
    if (m_dofs[2]) m_derGB[iStation+2*m_geometry.size()] = 1.;
    if (m_dofs[3]) m_derGB[iStation+3*m_geometry.size()] = Ymeas[iStation];
    if (m_dofs[4]) m_derGB[iStation+4*m_geometry.size()] = Xmeas[iStation];

    if (m_dofs[5]) m_derGB[iStation+5*m_geometry.size()] = Ymeas[iStation];

    if (m_dofs[0]) m_derNonLin[iStation]                     = 0.0;
    if (m_dofs[1]) m_derNonLin[iStation+m_geometry.size()]   = 0.0;
    if (m_dofs[2]) m_derNonLin[iStation+2*m_geometry.size()] = 1.0;  // Non linear param related to X track slope
    if (m_dofs[3]) m_derNonLin[iStation+3*m_geometry.size()] = 1.0;
    if (m_dofs[4]) m_derNonLin[iStation+4*m_geometry.size()] = 1.0;
    if (m_dofs[5]) m_derNonLin[iStation+5*m_geometry.size()] = 0.0;

    // And we put the X equation in Millepede

    m_alignTool.EquLoc(&m_derGB[0], &m_derLC[0], &m_derNonLin[0], Xmeas[iStation] , 
                     resolution); // Store hits parameters


    m_alignTool.ZerLoc(&m_derGB[0],&m_derLC[0],&m_derNonLin[0]); // Reset the derivatives arrays

    // Add the local derivatives for the Y part of the track

    m_derLC[0]= 0.0;
    m_derLC[1]= 0.0;
    m_derLC[2]= 1.0;
    m_derLC[3] = m_geometry[iStation].z();

    // Then add the global one

    if (m_dofs[0]) m_derGB[iStation]                     = 0.0;
    if (m_dofs[1]) m_derGB[iStation+m_geometry.size()]   = -1.0;

    // Linear approximation
    //    if (m_dofs[2]) m_derGB[iStation+2*m_geometry.size()] = 1.*misalignedTrack.ty();
    //    if (m_dofs[3]) m_derGB[iStation+3*m_geometry.size()] = Ymeas[iStation]*misalignedTrack->ty();
    //    if (m_dofs[4]) m_derGB[iStation+4*m_geometry.size()] = Xmeas[iStation]*misalignedTrack->ty();

    // Non-linear case
    if (m_dofs[2]) m_derGB[iStation+2*m_geometry.size()] = 1.;
    if (m_dofs[3]) m_derGB[iStation+3*m_geometry.size()] = Ymeas[iStation];
    if (m_dofs[4]) m_derGB[iStation+4*m_geometry.size()] = Xmeas[iStation];

    if (m_dofs[5]) m_derGB[iStation+5*m_geometry.size()] = -Xmeas[iStation];

    if (m_dofs[0]) m_derNonLin[iStation]                     = 0.0;
    if (m_dofs[1]) m_derNonLin[iStation+m_geometry.size()]   = 0.0;
    if (m_dofs[2]) m_derNonLin[iStation+2*m_geometry.size()] = 2.0;  // Non linear param related to Y track slope
    if (m_dofs[3]) m_derNonLin[iStation+3*m_geometry.size()] = 2.0;
    if (m_dofs[4]) m_derNonLin[iStation+4*m_geometry.size()] = 2.0;
    if (m_dofs[5]) m_derNonLin[iStation+5*m_geometry.size()] = 0.0;

    // And we put the Y equation in Millepede

    m_alignTool.EquLoc(&m_derGB[0], &m_derLC[0], &m_derNonLin[0], Ymeas[iStation] , 
		       resolution); // Store hits parameters
  }

  
  // local fit of the track
  std::vector<double> track_params;     // Vector containing the track parameters+errors 

  track_params.resize(2*m_derLC.size()); // Size is 2*n_local

  bool sc = m_alignTool.FitLoc(m_alignTool.GetTrackNumber(),&track_params[0],0);
  
  if (sc) m_alignTool.SetTrackNumber(m_alignTool.GetTrackNumber()+1);

  delete misalignedTrack;

  return true;
}


void MilleTool::reset() {m_nTrack = 0;}

bool MilleTool::globalFit(){

  // final global fit
  m_alignTool.MakeGlobalFit(&m_par[0],&m_error[0],&m_pull[0]);

  return true;
}

const std::vector<double>& MilleTool::parameters() const{
  return m_par;
}

const std::vector<double>& MilleTool::errors() const{
  return m_error;
}

const std::vector<double>& MilleTool::pulls() const{
  return m_pull;
}


